"""@author: Andrés Corrada-Emmanuel."""

# from ntqr import Labels
from ntqr.statistics import MClassifiersVariables

labels = ("ax", "by")
classifiers = ("ci", "cj")


def test_n1_r2_variables():
    vars = MClassifiersVariables(labels, classifiers[:1])



"""
Module with classes that encapsulate the production of statistical variables
for evaluation models.
"""

from itertools import combinations, product
from types import MappingProxyType
from typing_extensions import Iterable, Literal, Mapping, Sequence, Union

import sympy

import ntqr


class MClassifiersVariables:
    """
    Statistical responses variables for M classifiers.




    Attributes
    ----------

    qs : List[simpy.Symbol..]
        The Q_l_i variables. There are R of them.

    responses : Mapping[m_subset, Mapping[Sequence[Label], simpy.Symbol]]


    responses_by_label : Mapping[m_subset, Mapping[Label, Mapping[...]]]

    """

    def __init__(self, labels: ntqr.Labels, classifiers: Sequence[str]):
        """


        Parameters
        ----------
        labels : Labels
            Labels for the question responses.
        classifiers : Sequence[str]
            Labels for the classifiers.

        Returns
        -------
        None.

        """
        self.labels = labels
        self.classifiers = classifiers

        # The variables for the answer-key Q-simplex.
        self.qs = MappingProxyType(
            {label: sympy.Symbol(r"Q_" + label) for label in labels}
        )

        self._responses = {}
        # The observable responses by the classifiers
        for m in range(1, len(classifiers) + 1):
            for m_subset in combinations(classifiers, m):
                self._responses[m_subset] = self._response_variables(
                    labels, m_subset
                )
        self.responses = MappingProxyType(self._responses)
        del self._responses

        self._label_responses = {}
        # The observable responses by the classifiers
        for m in range(1, m + 1):
            for m_subset in combinations(classifiers, m):
                self._label_responses[m_subset] = (
                    self._label_response_variables(labels, m_subset)
                )
        self.label_responses = MappingProxyType(self._label_responses)
        del self._label_responses

    def _response_variables(self, labels, classifiers):
        """
        Constructs observable response variables given 'labels' and
        'classifiers'.

        Parameters
        ----------
        labels : List
            Labels to use.
        classifiers : Sequence[str]
            Labels to use for the classifiers. The label should support
            being stringified.

        Returns
        -------
        Dictionary of by-label response counts, one per label.
        """

        clsfr_strs = [str(classifier) for classifier in classifiers]
        vars = {}

        # The possible responses by N classifiers outputting R labels
        # a set of size R^N
        vars = {
            decisions: sympy.Symbol(
                r"R_{" + self.seq_str(decisions, clsfr_strs) + r"}"
            )
            for decisions in product(labels, repeat=len(classifiers))
        }
        return vars

    def _label_response_variables(self, labels, classifiers):
        """
        Constructs variables associated with correct and wrong
        response counts given true label.

        Parameters
        ----------
        labels : Sequence[Label]
            Labels to use.
        classifier : Sequence[str]
            Index of classifier.

        Returns
        -------
        Dictionary of by-label response counts, three per label.
        In addition, each label contains a 'correct' key
        that points to the variable associated with correct responses.
        An 'errors' dictionary is indexed by possible wrong
        label assignments.
        """

        vars = {
            true_label: {
                decisions: self.label_r_var_symbol(
                    decisions, classifiers, true_label
                )
                for decisions in product(labels, repeat=len(classifiers))
            }
            for true_label in labels
        }

        # Now we define the all correct variable and the
        # not-all-correct variables
        for true_label in labels:
            all_correct = tuple([true_label for classifier in classifiers])

            # The single variable corresponding to all of the classifiers
            # being correct on the true label.
            vars[true_label]["all_correct"] = self.label_r_var_symbol(
                all_correct,
                classifiers,
                true_label,
            )

            # The variables where at least one classifier is wrong
            vars[true_label]["errors"] = {
                decisions: self.label_r_var_symbol(
                    decisions, classifiers, true_label
                )
                for decisions in product(labels, repeat=len(classifiers))
                if decisions != all_correct
            }

        return vars

    def seq_str(self, decisions, classifiers) -> str:
        """


        Parameters
        ----------
        decisions : Sequence[labels]
            A sequence of N labels.
        clsfr_strs : TYPE
            The N classifier labels.

        Returns
        -------
        Comma separated string of the forms l_c.

        """
        return ",".join(
            [
                label + r"_{" + str(classifier) + r"}"
                for label, classifier in zip(decisions, classifiers)
            ]
        )

    def label_r_var_symbol(self, decisions, classifiers, label):
        return sympy.Symbol(
            r"R_{" + self.seq_str(decisions, classifiers) + r"," + label + r"}"
        )

    def r_var_symbol(self, decisions, classifiers):
        return sympy.Symbol(
            r"R_{" + self.seq_str(decisions, classifiers) + r"}"
        )

    def pair_correlations(
        self, pair: tuple[str, str], decisions: tuple[str, str], l_true: str
    ) -> sympy.UnevaluatedExpr:
        """


        Parameters
        ----------
        pair : tuple[str,str]
            Pair of classifiers.
        decisions : tuple[str,str]
            The decisions by the pair.

        Returns
        -------
        Mapping from label to an expression for the pair correlation
        for the decisions given true label.

        """
        q_l = self.qs[l_true]
        expr = (
            1
            / q_l
            * (
                self.label_responses[pair][l_true][decisions]
                - 1
                / q_l
                * self.label_responses[(pair[0],)][l_true][(decisions[0],)]
                * self.label_responses[(pair[1],)][l_true][(decisions[1],)]
            )
        )
        return expr

    def all_agree_subs_dict(
        self,
    ) -> Mapping[sympy.Symbol, sympy.UnevaluatedExpr]:
        """ """
        subs_dict = {}
        for m_subset, m_subset_dict in self.label_responses.items():
            for l_true in self.labels:
                var = m_subset_dict[l_true][
                    tuple([l_true for i in range(len(m_subset))])
                ]
                subs_dict[var] = self.qs[l_true] - sum(
                    m_subset_dict[l_true]["errors"].values()
                )

        return subs_dict


class SingleClassifierVariables:
    """Statistical variables associated with a single classifier.

    Attributes
    ----------

    question_numbers : List[simpy.Symbol..]
        The variables for the count of correct questions for each true
        label.

    responses : Mapping[label, simpy.Symbol]
        The variables for the observed interger count of labels in
        the test. Generally, of the form

        R_{l_ i} : Number of 'l' label responses by classifier 'i'

    responses_by_label: Mapping[label, simpy.Symbol]
        The variables associated with correct and wrong responses given
        the true label.

        Variables are of the form,
        R_{l_i, l_true} : Number of l responses by classifier 'i'
        given true label, l_true.

        Given true label, responses_by_label[label] returns a dictionary
        with keys:
            'correct': the variable for correct responses given true label.
            'errors': dictionary, indexed by wrong label, to incorrect
            responses.
            'l_1', 'l_2': dictionary, indexed by response label, given
            true_label.
    """

    def __init__(self, labels, classifier):
        "Constructs variables for 'labels'."

        self.questions_number = {
            label: sympy.Symbol(r"Q_" + label) for label in labels
        }
        self.responses = self.response_variables(labels, classifier)
        self.responses_by_label = self.label_response_variables(
            labels, classifier
        )

    def response_variables(self, labels, classifier):
        """
        Constructs observable response variables given 'labels' and
        'classifier'.

        Parameters
        ----------
        labels : List
            Labels to use.
        classifier : int
            Index of classifier.

        Returns
        -------
        Dictionary of by-label response counts, one per label.
        """

        clsfr_str = str(classifier)
        vars = {}
        for label in labels:
            vars[label] = sympy.Symbol(
                r"R_{" + label + r"_{" + clsfr_str + r"}}"
            )

        return vars

    def label_response_variables(self, labels, classifier):
        """
        Constructs variables associated with correct and wrong
        response counts given true label.

        Parameters
        ----------
        labels : List
            Labels to use.
        classifier : int
            Index of classifier.

        Returns
        -------
        Dictionary of by-label response counts, three per label.
        In addition, each label contains a 'correct' key
        that points to the variable associated with correct responses.
        An 'errors' dictionary is indexed by possible wrong
        label assignments.
        """

        clsfr_str = str(classifier)
        vars = {}
        for true_label in labels:
            label_vars = vars.setdefault(true_label, {})
            label_vars_error = label_vars.setdefault("errors", {})

            for response_label in labels:
                curr_var = sympy.Symbol(
                    r"R_{"
                    + response_label
                    + r"_{"
                    + clsfr_str
                    + r"},"
                    + true_label
                    + r"}"
                )
                label_vars[response_label] = curr_var

                # Now we add it to either 'correct' or 'errors'
                if response_label != true_label:
                    label_vars_error[response_label] = curr_var
                else:
                    label_vars["correct"] = curr_var

        return vars



"""@author: Andrés Corrada-Emmanuel."""

from typing import Iterable


class Label(str):
    """
    Label object to guarantee a label is stringifiable.

    """

    def __init__(self, label):
        """


        Parameters
        ----------
        label : str
            Label to appear in variable subscripts.

        Returns
        -------
        None.

        """
        self._label = label

    def __str__(self):
        return self._label


class Labels(tuple):
    """
    Labels used in test question responses. The NTQR package assumes that
    all test questions have the same, fixed, label set. These are a
    tuple-like object so the user can specify the canonical order of
    the labels.

    The number of labels defines the integer R in the acronym NTQR.
    Binary classification: R=2
    Three label classification: R=3, etc.
    """

    def __init__(self, labels: Iterable[str]):
        """


        Parameters
        ----------
        labels : Iterable[str]
           Labels for question responses.

        Returns
        -------
        None.

        """
        self._labels = tuple([Label(label) for label in labels])



"""
Module for classes and datastructures related to test sketches.

The logic of unsupervised evaluation can be viewed as similar to
data streaming algorithms. We use summary statistics of a test
(observable agreements and disagreements between classifiers)
to deduce other statistics of the test (the set of possible group
evaluations for them).
"""

from typing import Self
from typing_extensions import Mapping, Sequence
from itertools import combinations, product

from ntqr import Label, Labels


class QuestionAlignedDecisions:
    """
    The question aligned test sketch.

    It can be shown algebraically that the count of question aligned
    responses by the test takers are "generated" by statistics of
    their correctness on the test. These sample statistics of a given
    test include their individual average performance as well as
    sample statistics tracking their error correlations with each
    other. This coupling of observed test sketch and polynomial system
    of unknown sample statistics of test takers is the core idea of the
    NTQR package.

    This class represents the observable side of that
    coupling - the counts of the R^N ways that N test takers
    can agree and disagree when they are taking a test with
    R responses.
    """

    def __init__(
        self, observed_responses: Mapping[Sequence[Label], int], labels: Labels
    ):
        """
        Validate test responses and transform into question aligned counts.

        Parameters
        ----------
        observed_responses : Mapping[Sequence[Label], int]
            Observed counts of agreements and disagreements between
            N test takers responding to questions with R possible
            responses. This can contain at most R^N keys.
        labels : Labels
            The R labels.

        Returns
        -------
        None.

        """
        # It must have some keys
        if len([key for key in observed_responses.keys()]) == 0:
            raise ValueError("There must be at least one decision key.")

        # All the keys must be the same length
        key_lengths = list([len(key) for key in observed_responses.keys()])

        if not (len(set(key_lengths)) == 1):
            raise ValueError("Not all decision tuples have the same length.")

        # Fill up a new dictionary with AlignedDecisions keys
        self.N = len(list(observed_responses.keys())[0])

        if not all(
            [
                (decision in product(labels, repeat=self.N))
                for decision in observed_responses.keys()
            ]
        ):
            raise ValueError(
                """One or more decisions in 'observed_responses'
                contain label(s) not in 'labels' arg."""
            )

        self.counts = {
            decisions: observed_responses.get(decisions, 0)
            for decisions in product(labels, repeat=self.N)
        }

        # And remember the labels used in the decisions keys
        self.labels = labels

    def marginalize(self, indices: Sequence[int]) -> Self:
        """
        Marginalize counts to 'indices' only counts.

        Parameters
        ----------
        indices : Sequence[int]
            Subset of classifiers.

        Returns
        -------
        Self
            Question aligned counts for subset of classifiers specified
            by 'indices'.

        """
        new_counts = {}
        for key, count in self.counts.items():
            new_key = tuple([key[i] for i in indices])
            new_counts[new_key] = new_counts.setdefault(new_key, 0) + count

        return QuestionAlignedDecisions(new_counts, self.labels)

    def m_subset_indices_to_val(self, m: int) -> Mapping[tuple, int]:
        """
        Response counts for all possible classifier subsets of size 'm'.

        Parameters
        ----------
        m : int
            Size of the subsets of the aligned decisions to use.

        Returns
        -------
        Mapping from m-sized subsets indices to observed response counts.

        """
        if m > self.N:
            raise ValueError(
                "Size of m-subsets, m, cannot be larger than the number "
                + "of test-takers, N"
            )

        indices_to_val = {
            m_subset: self.marginalize(m_subset)
            for m_subset in combinations(range(self.N), m)
        }

        return indices_to_val



"""Module for super classes of R-axioms based evaluations.

Logically consistent evaluations are defined by the size of an ensemble
and its associated axioms.

"""

from collections.abc import Sequence
from dataclasses import dataclass, field
from itertools import chain, combinations, product
from types import MappingProxyType
from typing import Iterable, Mapping, Optional, Self, Set, Tuple

import numpy as np
import numpy.typing as npt
import sympy

from ntqr import Labels
from ntqr.algebra import extract_coefficents, extract_constant
from ntqr.testsketches import QuestionAlignedDecisions
from ntqr.statistics import MClassifiersVariables
from ntqr.raxioms import MAxiomsIdeal
import ntqr.r2.raxioms
import ntqr.r3.raxioms


class AnswerKeyQSimplex:
    """
    Class to generate all answer-key simplex points.

    Given the test size, Q, and the labels (R=r) of them. This
    is a set of tuples of dimension r, the number of labels.

    It exists in (R-1) space since, by construction, we
    must have their sum always equal to Q. The number of
    possible qs is 1/(R-1)!*(Q+1)*(Q+2)*...*(Q+R-1).

    To summarize:
        - The number of labels in the answer key is a tuple
        of size R. Its value is unknown in unsupervised settings.

        - The number of such tuples is 1/(R-1)!(Q+1)...

        - This set exists in a (R-1) dimensional space inside
          the R space for the tuples.

    """

    def __init__(self, Q: int, labels: Labels):
        """
        Class initializer.

        Parameters
        ----------
        Q : int
            The number of question in the test.
        labels : Labels
            The labels.

        Returns
        -------
        None.

        """
        self.Q = Q
        self.N = Q + 1

        # The code is not safe or reliable beyond R=3 in
        # the current version.
        self.labels = labels
        R = len(labels)
        self.R = R

    def qs(self) -> Iterable[tuple[int]]:
        """
        Generate all possible answer-key simplex points.

        Returns
        -------
        Iterable[tuple[int]]
            Generator of all possible answer-key simplex points.

        """
        ranges = (range(0, self.N) for label in self.labels)
        answer_key_simplex_points = filter(
            lambda x: sum(x) == self.Q, product(*ranges)
        )
        for point in answer_key_simplex_points:
            yield point


class MLabelResponseSimplexes:
    """
    Class to manage the response simplexes associated with each label.

    Each subset of sized-M for N classifiers has its own set of label
    response simplexes, one for each of the R labels in the test. These
    are the set of all possible values for statistics of aligned decisions
    by the members of the m-subset GIVEN true label.

    The a-priori logic of a test is that the sum of all possible responses
    by a m-subset must exactly equal to the Q_label, the count of the label
    in the answer key. This defines a simplex for a given label.

    The posterior logic is that the simplexes for any given value of M,
    M=m, have axioms that depend on all simplexes of value less than m.
    Thus the M=2 simplexes involve variables from the two M=1 simplexes
    that come from the classifier pair responses. The M=3 simplexes involve
    all M=2 simplexes, and all M=1 simplexes. And so on.

    This class is meant to internally manage that logical complexity
    by keeping track of all the variables that are needed for each
    simplex as well as the enclosing 'shells' that provide the values
    for the response variables.

    A separate class, MAxiomsVarieties, combines the functionality
    of this class and the ntqr.raxioms.MAxiomsIdeal class to compute
    the subset of possible label test responses that are logically
    consistent with the observed test results.

    """

    def __init__(
        self,
        labels: Sequence[str],
        classifiers: Sequence[str],
        responses: Mapping[tuple, int],
        qs,
        m,
    ):
        self.labels = labels
        self.qs = qs
        self.m = m
        self.classifiers = classifiers

        self.qad = QuestionAlignedDecisions(responses, labels)
        # We must have as many classifier labels as the length of
        # the response decision tuples.
        if len(classifiers) != self.qad.N:
            err_msg = (
                "There must be as many classifiers({}) as the length of the"
                + " response decision keys ({})."
            )
            raise ValueError(err_msg.format(len(classifiers), self.qad.N))
        # Currently only supporting up to m=1.
        if m > 3:
            raise NotImplementedError(
                "Only M=1,2,3 axioms supported currently."
            )

        self.qs = qs

        # Initialize dicts for vars and vals needed to define
        # the label simplexes for these classifiers
        self._initialize_response_dicts()

        self._initialize_sympy_response_dict()

    def _initialize_response_dicts(self):
        # We need to manage two sets of variables, the ones for the
        # observed test responses, let's call them 'responses', and
        # ones for those observed responses GIVEN true label. Call
        # those 'label_responses'
        # In addition, we want to set values for those vars. So
        # everything is organized by the decision tuples of an m-subset
        self.response_vars = {}
        # In an unsupervised setting we have access to the response values
        # after the test is taken.
        self.response_vals = {}
        # In unsupervised settings we only know these vars are needed.
        self.label_response_vars = {}
        #

        # Everything is organized by subsets of the classifier labels.
        # Then, the response vars then get pointed to by m-sized
        # decision tuples:
        #    m-subset -> m-decision tuple -> (var or val).
        # For label response vars:
        #    m-subset -> label -> m-decision tuples
        # We need all the 1-decision vars, the 2-decision vars,
        # up to m
        allMVars = MClassifiersVariables(self.labels, self.classifiers)
        for m_current in range(1, self.m + 1):
            m_response_vals = self.m_responses(m_current)
            for m_subset in combinations(self.classifiers, m_current):
                self.response_vars.update(
                    {m_subset: allMVars.responses[m_subset]}
                )
                self.label_response_vars.update(
                    {m_subset: allMVars.label_responses[m_subset]}
                )
                self.response_vals[m_subset] = m_response_vals[m_subset]

    def _initialize_sympy_response_dict(self):
        # Now we construct the evaluation dict needed by SymPy to evaluate
        # the axioms
        var_val_pairs = zip(
            [
                var
                for m_subset, m_decisions_dict in self.response_vars.items()
                for var in m_decisions_dict.values()
            ],
            [
                val
                for m_subset, m_decisions_dict in self.response_vals.items()
                for val in m_decisions_dict.values()
            ],
        )
        response_eval_dict = {var: val for var, val in var_val_pairs}
        # And add the q_s values
        self.q_vars = MappingProxyType(
            {label: sympy.Symbol(r"Q_" + label) for label in self.labels}
        )
        response_eval_dict.update(
            {
                self.q_vars[label]: val
                for label, val in zip(self.labels, self.qs)
            }
        )

        self.response_eval_dict = MappingProxyType(response_eval_dict)

        return

    def m_responses(self, m: int) -> Mapping[tuple, Mapping[tuple, int]]:
        """
        Observed responses by all m-sized subsets of the classifiers.

        Observed responses set the ceilings for any possible value for
        the label responses - responses given true label. For example,
        if we observe that two classifiers agreed on the same label some
        number of times, no possible evaluation of that agreement given
        true label can exceed this number.

        Observed responses create the 'ratchet' of evaluation. No label
        response variable can have a value larger than that of any subset
        of the classifiers responding similarly.

        Parameters
        ----------
        m : int
            Size of the subsets of the classifiers that will be used.

        Returns
        -------
        Mapping[tuple, Mapping[tuple, int]]
            This is semantically of the form,
                Mapping[m-subset-classifiers,
                    Mapping[m-decisions, observed count]
            A tuple that identifies the m-sized subset of N classifiers
            points to a mapping of question aligned decisions by that
            subset to the observed integer count in the test.

        """
        classifiers = self.classifiers
        # This code is somewhat confusing because of the convenience for
        # users to input test responses by a tuple ordering instead of
        # subscripting decisions. '(a,b)'  is easier to write than
        # set(a_i, b_j). It seems using the tuple is overall better
        # for immutability transparency in Python. Using set(...) would
        # not work and we would need immutable sets.
        # Hence, we define m_subsets by their indices into the decision
        # tuples in the user provided test responses.
        m_subsets = combinations(range(len(classifiers)), m)
        m_subsets_responses = self.qad.m_subset_indices_to_val(m)
        # And now we want to return a dictionary that maps from
        # the m-subset-indices to their decision tuple test counts.
        ret_val = {
            tuple([classifiers[i] for i in m_subset]): m_subsets_responses[
                m_subset
            ].counts
            for m_subset in m_subsets
        }

        return ret_val


@dataclass(frozen=True, kw_only=True)
class MVariety:
    """
    Class for the points obeying all axiom orders up to M.

    The M=1 variety contains points that obey the M=1 axioms
    for a single classifier. The M=2 variety contains points
    that obey the M=2 axioms for a pair of classifiers, but
    also the M=1 axioms for each of them.

    This class follows the 'code smell' test, it appears because
    we need operations that can create logically consistent
    intersections of varieties of fixed order.

    These varieties, by construction, do not use any information
    about responses at higher order. So the union of m=1 varieties
    never uses, or can contain, information about pair responses
    or higher.

    Thus the intersection of m-varities is a containing variety
    for the variety that corresponds to all axioms up to m=N
    being obeyed.
    """

    labels: tuple[str]
    classifiers: tuple[str]
    qs: Mapping[str, sympy.Symbol]
    m: int
    label_vars: tuple[tuple[sympy.Symbol]]
    points: Mapping[tuple, Mapping[tuple, {}]] = field(repr=False)

    def __eq__(self, other_variety: Self) -> bool:
        """
        Test equality.

        The current implementation is a weaker check on equality.
        It just verifies that self.m and self.label_vars are equal.
        If so, it returns true.

        A strict check on equality would verify that all points are
        also equal.

        Parameters
        ----------
        other_variety : Self
            Other variety.

        Returns
        -------
        bool
            Whether the varities are equal.

        """

        if self.m != other_variety.m:
            return False

        union_classifiers = self.union_classifiers(other_variety)
        other_only_vars = self.only_other_vars(other_variety)
        if (
            set(other_variety.classifiers).issubset(union_classifiers)
            and other_only_vars == tuple()
        ):
            return (True,)
        else:
            return False

    def __and__(self, other_variety: Self) -> Self:
        """
        Create logical intersection of self and other_variety.

        Constructing varieties of order M=m requires that we
        find the logically consistent intersection of m-1 varieties.
        Starting at M=3, two varieties of order m-1 share some of their
        variables. The 'and' operation returns the joined points that
        are logically consistent with each other: have the same value
        for the shared variables.

        Parameters
        ----------
        other_variety : MVariety
            Variety to do logical intersection with. It must be
            of the same M=m order as this one.

        Returns
        -------
        MVariety
            The points whose intersection is logically consistent.
        """
        # We first must make sure that the varieties are of the
        # same order: m_self = m_other_variety
        if not self.m == other_variety.m:
            raise ValueError(
                f"""Cannot perform logical intersection on varieties
                of different order. This variety is of order
                m={self.m}. 'other_variety' is of order m={other_variety.m}."""
            )

        if self == other_variety:
            return self

        return self.construct_intersection_variety(other_variety)

    def compute_from_indices(
        self, other_variety: Self, var_order: Sequence[sympy.Symbol]
    ) -> Sequence[int]:
        """
        Compute the indices for finding the values of the final variety.

        Whenever we do logical AND of varities, the final variety will contain
        values from both varieties. This function returns the indices in
        their concatenated variety point where we can find the variables
        specified in 'var_order.'

        Parameters
        ----------
        var_order : Sequence[sympy.Symbol]
            Sequence that defines the order of the variables.

        Returns
        -------
        Sequence[int]
            Source index in a concatenated variety point.

        """
        from_indices = {var_order[i]: 0 for i in range(len(var_order))}

        self_only_vars = self.only_self_vars(other_variety)
        all_self_vars = list(chain(*self.label_vars))
        in_self_index = {
            var: i for var, i in zip(all_self_vars, range(len(all_self_vars)))
        }

        other_only_vars = self.only_other_vars(other_variety)
        all_other_vars = list(chain(*other_variety.label_vars))
        in_other_index = {
            var: i
            for var, i in zip(all_other_vars, range(len(all_other_vars)))
        }

        common_vars = self.common_vars(other_variety)

        # Update where to find the data we need for each new point
        for var in self_only_vars:
            from_indices[var] = in_self_index[var]
        for var in common_vars:
            from_indices[var] = in_self_index[var]
        for var in other_only_vars:
            from_indices[var] = in_other_index[var] + len(all_self_vars)

        return from_indices

    def intersection_label_vars(
        self, other_variety: Self
    ) -> Set[sympy.Symbol]:
        """
        Find variables shared by both varieties.

        Beginning at m=2, varieties may share lower m variables. In addition,
        we want the __and__ operation to be idempotent when a previously
        joined variety is joined again - joining self with another variety
        gives the same result if we join it again.

        Parameters
        ----------
        other_variety : Self
            The other variety.

        Returns
        -------
        Set of shared variables between the two varieties.

        """
        return frozenset(set(self.label_vars) & set(other_variety.label_vars))

    def union_classifiers(self, other_variety: Self) -> Set[str]:
        """
        Find union of the classifiers in self and 'other_variety'.

        Parameters
        ----------
        other_variety : Self
            Variety to compare with.

        Returns
        -------
        Set[str]
            The union of the classifiers in self and 'other_variety'.

        """
        return tuple(
            sorted(set(self.classifiers) | set(other_variety.classifiers))
        )

    def var_order(self, other_variety: Self) -> tuple[sympy.Symbol]:
        """
        Construct var order of the intersection of self with 'other_variety'.

        Parameters
        ----------
        other_variety : Self
            A variety of the same order as self.

        Returns
        -------
        The var order for the points in the intersection of the varieties.

        """
        # We have to be careful, we will be missing variables
        vars_present = set(chain(*self.label_vars)).union(
            set(chain(*other_variety.label_vars))
        )
        union_classifiers = tuple(
            sorted(self.union_classifiers(other_variety))
        )
        mvars = MClassifiersVariables(self.labels, union_classifiers)
        vars = []
        for m_curr in range(1, self.m + 1):
            for m_subset in combinations(
                self.union_classifiers(other_variety), m_curr
            ):
                for label in self.labels:
                    for decisions in product(self.labels, repeat=m_curr):
                        var = mvars.label_responses[m_subset][label][decisions]
                        if var in vars_present:
                            vars.append(var)

        return tuple(vars)

    def common_m1_vars(
        self, other_variety: Self, var_order: tuple[sympy.Symbol]
    ) -> tuple[str]:
        """
        Find common m=1 vars following var order.

        Parameters
        ----------
        other_variety : Self
            Variety to check.
        var_order : tuple[sympy.Symbol]
            Canonical var order.

        Returns
        -------
        tuple[str]
            m=1 label responses variables shared by self with other_variety.

        """
        other_m1_vars = set(other_variety.label_vars[0])
        self_m1_vars = set(self.label_vars[0])
        intersection_set = self_m1_vars.intersection(other_m1_vars)
        common_m1_vars = [var for var in var_order if var in intersection_set]
        return common_m1_vars

    def m1_var_indices(
        self, m1_vars: Sequence[sympy.Symbol], variety: Self
    ) -> tuple[int]:
        """
        Find the position of m=1 vars in variety.label_vars.

        Parameters
        ----------
        m1_vars : Sequence[sympy.Symbol]
            m=1 label response variables for which we want the index.
        variety : Self
            The variety whose m1 variables will be indexed.

        Returns
        -------
        tuple[int]
            Indices of the m=1 label response variables in a variety point.

        """
        variety_m1_vars = variety.label_vars[0]
        indices = [
            i
            for i in range(len(variety_m1_vars))
            if variety_m1_vars[i] in m1_vars
        ]
        return indices

    def common_m2p_vars(
        self, other_variety: Self, var_order: tuple[sympy.Symbol]
    ) -> tuple[str]:
        """
        Find common m=2 or higher vars following var order.

        Parameters
        ----------
        other_variety : Self
            Other variety.
        var_order : tuple[sympy.Symbol]
            Canonical var order.

        Returns
        -------
        tuple[str]
            m>=2 label response variables shared by self and 'other_variety'.

        """
        other_m2p_vars = set(chain(*other_variety.label_vars[1:]))
        self_m2p_vars = set(chain(*self.label_vars[1:]))
        intersection_set = self_m2p_vars.intersection(other_m2p_vars)
        common_m2p_vars = [var for var in var_order if var in intersection_set]
        return common_m2p_vars

    def m2p_var_indices(
        self, m2p_vars: Sequence[sympy.Symbol], variety: Self
    ) -> tuple[int]:
        """
        Find the indices for the m>=2 label responses variables.

        Parameters
        ----------
        m2p_vars : Sequence[sympy.Symbol]
            m>=2 label responses variables.
        other_variety : Self
            The other variety.

        Returns
        -------
        tuple[int]
            Indices of the m>=2 label response variables in a point from
            'variety'.

        """
        these_m2p_vars = list(chain(*variety.label_vars[1:]))
        indices = [
            i
            for i in range(len(these_m2p_vars))
            if these_m2p_vars[i] in m2p_vars
        ]
        return indices

    def only_self_vars(self, other_variety: Self) -> tuple[sympy.Symbol]:
        """
        Find label response variables only self has.

        Parameters
        ----------
        other_variety : Self
            The other variety.

        Returns
        -------
        tuple[sympy.Symbol]
        Label response variables only found in self and not in 'other_variety'.

        """
        only_in_self = tuple(
            var
            for var in list(chain(*self.label_vars))
            if var
            in set(chain(*self.label_vars)).difference(
                set(chain(*other_variety.label_vars))
            )
        )
        return only_in_self

    def only_other_vars(self, other_variety: Self) -> tuple[sympy.Symbol]:
        """
        Find label response variables only 'other_variety' has.

        Parameters
        ----------
        other_variety : Self
            The variety we are comparing with.

        Returns
        -------
        tuple[sympy.Symbol]
        Label response variables only found in 'other_variety'.

        """
        only_in_other = tuple(
            var
            for var in list(chain(*other_variety.label_vars))
            if var
            in set(chain(*other_variety.label_vars)).difference(
                set(chain(*self.label_vars))
            )
        )
        return only_in_other

    def common_vars(self, other_variety: Self) -> tuple[sympy.Symbol]:
        """
        Find label response variables in self and other_variety.

        Parameters
        ----------
        other_variety : Self
            The other variety.

        Returns
        -------
        tuple[sympy.Symbol]
        Label response variables found in self and other_variety.

        """
        return tuple(
            var
            for var in list(chain(*self.label_vars))
            if var
            in set(chain(*self.label_vars)).intersection(
                set(chain(*other_variety.label_vars))
            )
        )

    def var_indices(
        self, vars: Sequence[sympy.Symbol], variety: Self
    ) -> Iterable:
        """
        Get indices for vars in self.label_vars.

        Parameters
        ----------
        vars : Sequence[sympy.Symbol]
            Variables for which we want indices in points from 'variety'.
        variety : Self
            The variety where the indices are to be found.

        Returns
        -------
        Iterable
            The indices for 'vars' in a point from 'variety'.

        """
        var_indices = tuple(
            i
            for i in range(len(variety.label_vars))
            if variety.label_vars[i] in vars
        )

        return var_indices

    def join_label_vars(
        self, var_order: Sequence[sympy.Symbol], other_variety: Self
    ) -> tuple[tuple[sympy.Symbol]]:
        """
        Join label_vars by m-order.

        Parameters
        ----------
        other_variety : Self
            The variety whose variables will be joined with self.

        Returns
        -------
        The joined vars by m-order.

        """
        joined_vars = tuple(
            tuple(
                var
                for var in var_order
                if var in set(self_vars).union(set(other_vars))
            )
            for self_vars, other_vars in zip(
                self.label_vars, other_variety.label_vars
            )
        )

        return joined_vars


class MVarietyTupleDict(MVariety):
    """Concrete class for MVariety storing points as dict of tuples.

    Warning: this class is memory intensive.
    """

    def generate_points(self) -> Iterable:
        """
        Generate the points in this variety.

        Yields
        ------
        Iterable
            Points in the variety.

        """

        for key in self.points.keys():
            list_val = self.points[key]
            for val in list_val:
                yield key + val

    def generate_consistent_point_pairs(
        self, var_order: Sequence[sympy.Symbol], other_variety: Self
    ) -> Iterable[tuple[tuple[int], tuple[int]]]:
        """
        Generate pairs of points from each variety consisten with each other.

        Whenever we are joining varieties of order m >= 2, care must
        be taken to only combine points that agree on their common variables.
        For example, if we have the variety for classifiers 'i' and 'j' and
        want to join it with the variety for classifiers 'j' and 'k', we
        must make sure that we return point pairs, one from each variety,
        that agree in their values of 'j' responses.

        Parameters
        ----------
        var_order : Sequence[sympy.Symbol]
            Sequence of vars that defines order of variables in the
            joined variety.
        other_variety : Self
            The variety to check consistency on common variables.

        Yields
        ------
        (Iterable[tuple[tuple[int], tuple[int]]])
            Pairs of points, one from each variety, that are logically
            consistent with each other (agree on common variables).

        """
        common_m2p_vars = self.common_m2p_vars(other_variety, var_order)
        scm2pi = self.m2p_var_indices(common_m2p_vars, self)
        ocm2pi = self.m2p_var_indices(common_m2p_vars, other_variety)

        common_m1_vars = self.common_m1_vars(other_variety, var_order)

        for self_key, other_key in self.generate_consistent_key_pairs(
            common_m1_vars, other_variety
        ):

            for self_tail, other_tail in self.generate_consistent_tail_pairs(
                other_variety, self_key, other_key, scm2pi, ocm2pi
            ):

                yield (self_key + self_tail, other_key + other_tail)

    def generate_consistent_key_pairs(
        self,
        common_m1_vars: Sequence[sympy.Symbol],
        other_variety: Self,
    ) -> Iterable[Tuple[Tuple[int], Tuple[int]]]:
        """
        Generate key pairs for self and other_variety that are consistent.

        Parameters
        ----------
        other_variety : Self
            The variety we are joining self with.

        Returns
        -------
        Tuples of m1 keys for the points dict of each variety.

        """
        # Self common m1 var indices
        scm1i = self.m1_var_indices(common_m1_vars, self)
        ocm1i = self.m1_var_indices(common_m1_vars, other_variety)
        for self_key, other_key in product(
            self.points.keys(), other_variety.points.keys()
        ):
            test_value = [self_key[i] for i in scm1i] == [
                other_key[i] for i in ocm1i
            ]

            if test_value:
                yield (self_key, other_key)

    def generate_consistent_tail_pairs(
        self,
        other_variety: Self,
        self_key: Tuple[int],
        other_key: Tuple[int],
        scm2pi: Tuple[int],
        ocm2pi: Tuple[int],
    ) -> Iterable[Tuple[Tuple[int], Tuple[int]]]:
        """
        Generate key pairs for self and other_variety that are consistent.

        Parameters
        ----------
        other_variety : Self
            The variety we are joining self with.

        Returns
        -------
        Tuples of m1 keys for the points dict of each variety.

        """
        if self.m == 1:
            yield (tuple(), tuple())
        for self_tail, other_tail in product(
            self.points[self_key],
            other_variety.points[other_key],
        ):
            test_value = [self_tail[i] for i in scm2pi] == [
                other_tail[i] for i in ocm2pi
            ]
            if test_value:
                yield (self_tail, other_tail)

    def construct_intersection_variety(self, other_variety: MVariety) -> Self:
        """
        Construct the intersection.

        Parameters
        ----------
        other_variety : MVariety
            The variety we are intersecting self with.

        Returns
        -------
        Self
            The intersection of the two varieties..

        """
        # We first construct the var order in the final variety
        var_order = self.var_order(other_variety)

        # Figure out where to find the data needed for the final variety
        from_indices = self.compute_from_indices(other_variety, var_order)

        # We want to create the key so we need the length of the m=1
        # vars
        joined_label_vars = self.join_label_vars(var_order, other_variety)
        len_key = len(joined_label_vars[0])
        points_dict = {}

        var_indices = [from_indices[var] for var in var_order]

        for self_point, other_point in self.generate_consistent_point_pairs(
            var_order, other_variety
        ):
            cat_point = self_point + other_point
            new_key = tuple(cat_point[i] for i in var_indices[:len_key])
            new_tail = tuple(cat_point[i] for i in var_indices[len_key:])
            key_list = points_dict.setdefault(new_key, [])
            key_list.append(new_tail)

        return MVarietyTupleDict(
            labels=self.labels,
            classifiers=self.union_classifiers(other_variety),
            qs=self.qs,
            m=self.m,
            label_vars=joined_label_vars,
            points=points_dict,
        )


class MAxiomsVarieties:
    """
    Class to compute the test evaluation variety.

    The test evaluation variety for M=m axioms is the set of
    evaluations that are logically consistent with how we observe
    the classifiers agreeing and disagreeing on the question
    responses.
    """

    def __init__(
        self,
        labels: Sequence[str],
        classifiers: Sequence[str],
        responses: Mapping[tuple, int],
        qs,
        m,
    ):
        self.labels = labels
        self.classifiers = classifiers
        self.qs = qs
        self.m = m

        self.r_simplexes = MLabelResponseSimplexes(
            labels, classifiers, responses, qs, m
        )

        # Since we have the responses and the value of qs on the
        # Q-simplex, we can construct all the axioms needed
        # to define the evaluation ideals.

        # We instantiate the test axioms at this qs. This will
        # make them linear equations that only contain label
        # response variables.
        self.test_axioms = MappingProxyType(self.instantiate_axioms(m))

        self.mm1_relevant_vars = self._relevant_vars()
        self.var_max_responses = self._var_max_responses()

    def instantiate_axioms(self, m: int):
        """
        Fill in all response variables in the test axioms.

        Any m-axiom contains three sets of variables:
            1. The answer-key simplex variables. This function
               fills them in with the values given during
               class instantiation.
            2. The observed response counts. This function fills
               them in.
            3. All label m or less response variables. These need
               to be filled in as we move up the ladder of logical
               consistency.

        Parameters
        ----------
        m : int
            The order of the axioms, an integer of value 1 or greather.

        Returns
        -------
        Mapping[m_subset:instantiated_m_axioms]

        """
        response_eval_dict = self.r_simplexes.response_eval_dict

        axioms = {
            m_subset: {
                label: sympy.simplify(
                    MAxiomsIdeal(self.labels, m_subset, m_current)
                    .m_complex[m_subset]["axioms"][label]
                    .subs(response_eval_dict)
                )
                for label in self.labels
            }
            for m_current in range(1, m + 1)
            for m_subset in combinations(self.classifiers, m_current)
        }

        return axioms

    def _var_max_responses(self) -> Mapping[sympy.Symbol, int]:
        """
        Calculate the maximum possible value of all label response vars.

        Returns
        -------
        Mapping[sympy.Symbol, int]
        The observable value for the label response variable.

        """
        label_responses = self.r_simplexes.label_response_vars
        max_response_possible = {}
        for m_curr in range(1, self.m + 1):
            for m_subset in combinations(self.classifiers, m_curr):
                for decisions in product(self.labels, repeat=len(m_subset)):
                    for l_true in self.labels:
                        var = label_responses[m_subset][l_true][decisions]
                        max_response_possible[var] = (
                            self.r_simplexes.response_vals[m_subset][decisions]
                        )

        return max_response_possible

    def _relevant_vars(self):
        # Data structure needed to connect m response variables to
        # their relevant m-1 counterparts.
        label_responses = self.r_simplexes.label_response_vars
        relevant_vars = {}
        for m_curr in range(1, self.m + 1):
            for m_subset in combinations(self.classifiers, m_curr):
                m_vars = relevant_vars.setdefault(m_subset, {})
                for decisions in product(self.labels, repeat=len(m_subset)):
                    decision_vars = m_vars.setdefault(decisions, {})
                    for l_true in self.labels:
                        label_vars = decision_vars.setdefault(l_true, [])
                        label_vars.append(
                            label_responses[m_subset][l_true][decisions]
                        )

        mm1_relevant_vars = {}

        # A mapping between an m variable and m-1 variables that
        # constrain its max
        for m_curr in range(2, self.m + 1):
            for m_subset in combinations(self.classifiers, m_curr):
                for decisions in product(self.labels, repeat=m_curr):
                    for l_true in self.labels:
                        # What label response variable are we using?
                        var = label_responses[m_subset][l_true][decisions]
                        curr_list = mm1_relevant_vars.setdefault(var, [])
                        for i_subset in combinations(
                            range(m_curr), m_curr - 1
                        ):
                            for sub_clsfs, sub_decisions in zip(
                                tuple(m_subset[i] for i in i_subset),
                                tuple(decisions[i] for i in i_subset),
                            ):

                                curr_list.extend(
                                    relevant_vars[tuple(sub_clsfs)][
                                        tuple(sub_decisions)
                                    ][l_true]
                                )

        return mm1_relevant_vars

    def simplex_points_equal(self, total: int, maxs: Sequence[int], N: int):
        """
        Generate all simplex points with values less than or equal to maxs.

        This is a recursive generator to handle arbitrary number of variables
        in a simplex.

        Parameters
        ----------
        total : int
            Total value required for the sum of the vars on the simplex.
        maxs : Sequence[int]
            The max integer value that a var can have on that simplex.
        N : int
            The number of vars defining the simplex.

        Yields
        ------
        tuple[int]
            Simplex points, of dimension 'N', that sum to 'total' and whose
            var values do not exceed maxs values.

        """
        if sum(maxs) < total:
            return
        if N == 1:
            if maxs[0] > total:
                yield [total]
            else:
                yield [maxs[0]]
        else:
            for x in range(min(maxs[0], total) + 1):
                for part in self.simplex_points_equal(
                    total - x, maxs[1:], N - 1
                ):
                    yield [x] + part

    def mvariety(self, classifiers: Sequence[str]) -> MVariety:
        """
        Construct the MVariety dataclass for these classifiers.

        Parameters
        ----------
        classifiers : Sequence[str]
            The classifiers.

        Returns
        -------
        MVariety
            Dataclass containing the points in the variety.

        """
        m = len(classifiers)
        # print("Doing m variety: ", m)

        # We concatenate all the label response vars sorted by
        # self.labels
        labels_mvars = [
            [
                self.r_simplexes.label_response_vars[classifiers][l_true][
                    decisions
                ]
                for decisions in product(self.labels, repeat=len(classifiers))
            ]
            for l_true in self.labels
        ]

        labels_vars = list(chain(*labels_mvars))

        mvars_axioms_coeffs = self.turn_axiom_exprs_to_vectors(
            classifiers, labels_vars
        )

        # Root case
        if m == 1:
            # We are faking the empty set to keep the function call
            # the same for turn_axioms_exprs_to_linear_vectors
            mm1_point = np.empty(0, dtype=np.uint16)
            # We have r of these label simplexes
            labels_m_simplexes = (
                self.label_msimplex(label_mvars, {}, ql, mm1_point)
                for label_mvars, ql in zip(labels_mvars, self.qs)
            )
            # A point in this space is a tuple of r label simplexes,
            # one for each true label.
            variety_points = [
                mpoint
                for mpoint in product(*labels_m_simplexes)
                if self.satisfies_axioms(mpoint, mvars_axioms_coeffs)
            ]

            return MVarietyTupleDict(
                labels=self.labels,
                classifiers=classifiers,
                qs=self.qs,
                m=m,
                label_vars=(tuple(labels_vars),),
                points={tuple(sum(point, [])): [] for point in variety_points},
            )

        elif m > 1:
            mm1_varieties = [
                self.mvariety(mm1_subset)
                for mm1_subset in combinations(classifiers, m - 1)
            ]
            # We now consider the m-simplex points. This logic is
            # currently not for general m, it assumes m <= 2
            # at m=2 the mm1 varieties are disjoint between the
            # classifiers - sets of distinct size 1 are disjoint.
            # But distinct pairs may have a classifier in common
            # and so on.

            # We need the order of the vars in the points of each variety
            # And we need to sort them for reconciliation
            mm1_joined_variety = mm1_varieties[0]
            for mm1_variety in mm1_varieties:
                mm1_joined_variety &= mm1_variety
            mm1_vars = tuple(chain(*mm1_joined_variety.label_vars))
            mm1_indices = {
                var: i for var, i in zip(mm1_vars, range(len(mm1_vars)))
            }

            # We drop the constant term for the mm1 coeffs vector
            mm1_coeffs = [
                coeffs[1]
                for coeffs in self.turn_axiom_exprs_to_vectors(
                    classifiers, mm1_vars
                )
            ]

            points_dict = {}
            for mm1_point in mm1_joined_variety.generate_points():

                # Calculate the contribution of the mm1 vars to the axioms
                new_coeffs = []
                for label_mcoeffs, label_mm1coeffs in zip(
                    mvars_axioms_coeffs, mm1_coeffs
                ):
                    curr_coeff = (
                        label_mcoeffs[0]
                        + np.dot(np.array(mm1_point), label_mm1coeffs),
                        label_mcoeffs[1],
                    )
                    new_coeffs.append(curr_coeff)

                # We have r of these label simplexes
                labels_m_simplexes = (
                    self.label_msimplex(
                        label_mvars, mm1_indices, ql, mm1_point
                    )
                    for label_mvars, ql in zip(labels_mvars, self.qs)
                )
                # A point in this space is a tuple of r label simplexes,
                # one for each true label.
                variety_points = [
                    mpoint
                    for mpoint in product(*labels_m_simplexes)
                    if all(
                        (
                            (
                                coeffs[0]
                                + np.dot(
                                    coeffs[1], np.array(list(chain(*mpoint)))
                                )
                            )
                            == 0
                        )
                        for coeffs in new_coeffs
                    )
                ]

                # The indices of the key are all the values
                # at the front that point to a point in single
                # classifier responses
                len_key = len(classifiers) * (len(self.labels)) ** 2

                for point in variety_points:
                    new_point = mm1_point + tuple(chain(*point))
                    point_list = points_dict.setdefault(
                        new_point[:len_key], []
                    )
                    point_list.append(new_point[len_key:])

            header_vars = list(mm1_joined_variety.label_vars)
            header_vars.append(tuple(labels_vars))

            return MVarietyTupleDict(
                labels=self.labels,
                classifiers=classifiers,
                qs=self.qs,
                m=m,
                label_vars=header_vars,
                points=points_dict,
            )

    def turn_axiom_exprs_to_vectors(
        self, classifiers: Sequence[str], labels_vars: Sequence[sympy.Symbol]
    ) -> tuple[npt.NDArray[np.int16]]:
        """
        Turn label axioms into an array of coefficient vectors.

        Parameters
        ----------
        classifiers : Sequence[str]
            Sequence of the classifiers to consider.
        labels_vars : Sequence[sympy.Symbol]
            The order of the label vars in the returned tuple of ints.


        Returns
        -------
        var_coefficients : tuple[npt.NDArray[np.int16]]
            Sequence of integer coefficients for the axioms, one for each
            label.
        """
        test_axioms = self.test_axioms[classifiers]
        var_coefficients = []

        for axiom in test_axioms.values():
            curr_vec = (
                extract_constant(axiom),
                self.label_coefficients(labels_vars, axiom),
            )
            var_coefficients.append(curr_vec)

        return var_coefficients

    def label_coefficients(
        self, labels_vars: Sequence[sympy.Symbol], axiom: sympy.UnevaluatedExpr
    ) -> tuple[int]:
        """
        Compute the coefficients for labels_vars than appear in axiom.

        Parameters
        ----------
        labels_vars : Sequence[sympy.Symbol]
            Variables for which we want coefficients in axiom.
        axiom : sympy.UnevaluatedExpr
            The algebraic expression of the axiom.

        Returns
        -------
        tuple[int]
            The integer coefficients for label_vars in axiom.

        """
        coeffs = extract_coefficents(axiom, labels_vars)

        return np.array(coeffs, dtype=np.int16)

    def label_msimplex(
        self,
        label_mvars: Sequence[sympy.Symbol],
        var_indices: Mapping[sympy.Symbol, int],
        ql: int,
        mm1_point: npt.NDArray[np.uint16],
    ) -> Sequence[npt.NDArray[np.uint16]]:
        """
        Generate all label m-response vars points.

        Each label has an m-response simplex that must
        be logically consistent with lower m response varieties.

        Parameters
        ----------
        classifiers : Sequence[str]
            The classifiers for this m-response simplex.
        label : str
            The true label.
        ql : int
            Assumed count of the true label in the answer key.
        mm1_point : Iteratable[dict]
            The m (m-1)-varieties that define the starting point
            for creating the m-variety of these classifiers.

        Returns
        -------
        Generator of all points on the label m-response simplex.

        """
        maxs = self.vars_max_values(label_mvars, var_indices, ql, mm1_point)

        for point in self.simplex_points_equal(ql, maxs, len(maxs)):
            yield point

    def vars_max_values(
        self,
        label_mvars: Sequence[sympy.Symbol],
        var_indices: Mapping[sympy.Symbol, int],
        ql: int,
        mm1_point: npt.NDArray[np.uint16],
    ) -> Sequence[int]:
        """
        Calculate the 'ratchet of crowd evaluation'.

        Every label response var is a positive integer between zero and,
        at most, the Q_label assumed value. However, this max is most
        of the time larger than the logically consistent solutions.

        The max integer value is the minimum of:
            1. Q_label
            2. R_decisions
            3. all R_decisions_label for the m (m-1)-sized subsets of
               the classifier decisions.

        Clearly the max of label response variables must be Q_label. But
        if we observe the classifiers collectively producing a lower count,
        then that is the max. For example, if we never observed a pair count
        for (l_1, l_2) then all label response vars for this tuple must be
        zero for all labels.

        The same applies to the (m-1)-subsets of the decisions tuple given
        the true label. No value of the response count for the m-sized
        decisions tuple can be higher than any response count for the
        (m-1)-sized subsets of the decisions.

        Parameters
        ----------
        classifiers : Sequence[str]
            The m classifiers.
        decisions : Sequence[str]
            Their m-decisions.
        label : str
            The true label.
        mm1_point : dict
            The values of the mm1 variables.

        Returns
        -------
        int
            Max integer value for these classifier decisions given
            true label and the m-1 varieties point.

        """
        # The m-1 vars must be turned into values
        max_responses = [self.var_max_responses[var] for var in label_mvars]
        # We also cannot exceed the mm1_point values.
        if len(mm1_point) > 0:
            relevant_vars = [
                self.mm1_relevant_vars[var] for var in label_mvars
            ]
            mm1_point_indices = [
                [var_indices[var] for var in rel_vars]
                for rel_vars in relevant_vars
            ]
            mm1_mins = [
                min(mm1_point[i] for i in indices)
                for indices in mm1_point_indices
            ]

            return [
                min(ql, res_val, mm1_val)
                for res_val, mm1_val in zip(max_responses, mm1_mins)
            ]
        else:
            return [min(ql, max_val) for max_val in max_responses]

    def make_consistent_if_possible(
        self, mm1_points: Sequence[npt.NDArray[np.uint16]]
    ) -> npt.NDArray[np.uint16]:
        """
        Make a consistent m point or return an empty numpy array.

        Parameters
        ----------
        mm1_points : Sequence[npt.NDArray[np.uint16]]
            Order m minus 1 points to be joined logically.

        Returns
        -------
        npt.NDArray[np.uint16]
            An array of the joined points if possible, empty otherwise.

        """
        # Again, this function will not work correctly for m = 2 varieties
        # or higher.
        # We concatenate the point arrays for now
        consistent_point = np.concatenate(mm1_points).tolist()
        return consistent_point

    def satisfies_axioms(
        self,
        mpoint: tuple[tuple[int]],
        maxioms: Sequence[tuple[int, npt.NDArray[np.uint16]]],
    ) -> bool:
        """
        Test if mpoint satisfies maxioms.

        The variety is defined by all those points in the label
        response simplexes that satisfy the m-axioms.

        Parameters
        ----------
        classifiers : Sequence[str]
            The classifiers being considered.
        mpoint : Iterable[dict]
            Iteratable of label response point dicts.
        maxioms :  Mapping[str, tuple]
            The m-axioms reduced the the components for a linear operation.

        Returns
        -------
        bool
            Does this order m point obey the m-order axioms?

        """
        return all(
            (
                (
                    coeffs[0]
                    + np.dot(coeffs[1], np.array(np.concatenate(mpoint)))
                )
                == 0
            )
            for coeffs in maxioms
        )


class SingleClassifierEvaluations:
    """
    Class for the evaluations of a single classifier given test responses.

    The axioms of unsupervised evaluation are algebraic filters
    that narrow the set of possible evaluations for test takers.
    This filtering proceeds in a ladder-like manner. The single
    classifier axioms are the bottom rung of that filtering.

    For a test with R labels, the set of possible evaluations
    is a collection of
    has dimension R*(R-1) at each setting of the number of label
    correct in the answer key. This is the space of error responses.
    There being R-1 for each of the R labels. The single classifier
    axioms cut that dimension
    by R-1. So the set of evaluations for a single classifier consists
    of points that have dimension R*(R-1), but the set itself has
    dimension (R-1)*(R-1).


    """

    def __init__(
        self,
        Q: int,
        single_axioms: (
            ntqr.r2.raxioms.SingleClassifierAxioms
            | ntqr.r3.raxioms.SingleClassifierAxioms
        ),
    ):
        self.Q = Q
        self.axioms = single_axioms

    def correct_at_qs(
        self, qs: Sequence[int], responses: Sequence[int]
    ) -> set[Sequence[int]]:
        """Calculate all possible correct responses given qs.

        Parameters
        ----------
        qs : Sequence[int]
            Number of label questions.
        responses : Sequence[int]
            Label responses by classifier.

        Returns
        -------
        set[Sequence[int]]
            Set of possible label correct evaluations given qs and responses.
        """
        errors_at_qs = self.errors_at_qs(qs, responses)
        return set(
            ((ql - sum(label_errors)) for ql, label_errors in zip(qs, errors))
            for errors in errors_at_qs
        )

    def _check_axiom_consistency_(self, eval_dict, wrong_vars, wrong_vals):
        eval_dict.update(
            {var: val for var, val in zip(wrong_vars, wrong_vals)}
        )
        return self.axioms.satisfies_axioms(eval_dict)



"""@author: Andrés Corrada-Emmanuel."""
import itertools

import pytest

from ntqr import Label, Labels

from ntqr.evaluations import QuestionAlignedDecisions
from ntqr.r2.examples import uciadult_label_counts

#
# Testing R=2 QuestionAlignedDecisions functionality
#


@pytest.mark.parametrize(
    "observed_responses, error_msg",
    (
        ({}, "There must be at least one decision key."),
        ({("a", "b"): 1, ("a",): 3}, ""),
    ),
)
def test_questionaligneddecisions_valueerror(observed_responses, error_msg):
    with pytest.raises(ValueError, match=error_msg):
        QuestionAlignedDecisions(observed_responses, ("a", "b"))


#
# Testing marginalization of QuestionAlignedDecisions
#
labels = ("a", "b")
observed_responses = {}
for label, decisions_counts in uciadult_label_counts.items():
    for decisions, count in decisions_counts.items():
        observed_responses[decisions] = (
            observed_responses.get(decisions, 0) + count
        )


@pytest.mark.parametrize(
    "indices, counts",
    (
        ((0,), {("a",): 15389, ("b",): 21453}),
        ((1,), {("a",): 2671, ("b",): 34171}),
        ((2,), {("a",): 15061, ("b",): 21781}),
    ),
)
def test_questionaligneddecisions_marginalize(indices, counts):
    qads = QuestionAlignedDecisions(observed_responses, labels)
    assert qads.marginalize(indices).counts == counts



"""@author: Andrés Corrada-Emmanuel."""

import ntqr.evaluations

# This is the most important test module for the NTQR package.
# The current tests use experimental data for binary classifiers.
# A better test suite should include at least R=3 tests since that
# is when two classifiers can disagree and both be wrong.

# This test sketch comes from experiments following Bin Yu's Chapter
# 13 Jupyter notebook experiments with the online consumer dataset.
vote_counts = {
    ("a", "a", "a"): 31,
    ("a", "a", "b"): 8,
    ("a", "b", "a"): 27,
    ("a", "b", "b"): 52,
    ("b", "a", "a"): 1,
    ("b", "a", "b"): 1,
    ("b", "b", "a"): 8,
    ("b", "b", "b"): 72,
}

qa = 195
mVarieties = ntqr.evaluations.MAxiomsVarieties(
    ("a", "b"), ("i", "j", "k"), vote_counts, (qa, 200 - qa), 2
)
ijVariety = mVarieties.mvariety(
    (
        "i",
        "j",
    )
)
ikVariety = mVarieties.mvariety(
    (
        "i",
        "k",
    )
)
jkVariety = mVarieties.mvariety(
    (
        "j",
        "k",
    )
)
iVariety = mVarieties.mvariety(("i",))
jVariety = mVarieties.mvariety(("j",))
kVariety = mVarieties.mvariety(("k",))



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from pathlib import Path
from ntqr import __version__ as ntqr_version


sys.path.insert(0, str(Path("..", "..", "src").resolve()))
# sys.path.insert(0, str(Path("..", "src").resolve()))
# sys.path.insert(0, str(Path("..").resolve()))


# -- Project information -----------------------------------------------------

project = "NTQR"
copyright = "2024, Andrés Corrada-Emmanuel, Walker Lee, Adam Sloat"
authors = ["Andrés Corrada-Emmanuel", "Walker Lee", "Adam Sloat"]

# The full version, including alpha/beta/rc tags
version = ntqr_version
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    # Using myst-nb automatically imports myst-parser,
    # "myst_nb",
    "myst_parser",
    "nbsphinx",
    "sphinx.ext.mathjax",
    "autoapi.extension",
    "sphinx.ext.napoleon",
]

myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_image",
]
myst_url_schemes = ("http", "https", "mailto")

# myst_enable_extensions = [
#     "amsmath",
#     "attrs_inline",
#     "colon_fence",
#     "deflist",
#     "dollarmath",
#     "fieldlist",
#     "html_admonition",
#     "html_image",
#     "replacements",
#     "smartquotes",
#     "strikethrough",
#     "substitution",
#     "tasklist",
# ]

# Don't add .txt suffix to source files:
html_sourcelink_suffix = ""

# List of arguments to be passed to the kernel that executes the notebooks:
# nbsphinx_execute_arguments = [
#     "--InlineBackend.figure_formats={'svg', 'pdf'}",
# ]

# nbsphinx_thumbnails = {
#     "gallery/thumbnail-from-conf-py": "gallery/a-local-file.png",
#     "gallery/*-rst": "images/notebook_icon.png",
#     "orphan": "_static/favicon.svg",
# }

# nb_ipywidgets_js = {
#     "https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.4/require.min.js": {
#         "integrity": "sha256-Ae2Vz/4ePdIu6ZyI/5ZGsYnb+m0JlOmKPjt6XZ9JJkA=",
#         "crossorigin": "anonymous",
#     },
#     "https://cdn.jsdelivr.net/npm/@jupyter-widgets/html-manager@1.0.6/dist/embed-amd.js": {
#         "data-jupyter-widgets-cdn": "https://cdn.jsdelivr.net/npm/",
#         "crossorigin": "anonymous",
#     },
# }


nbsphinx_assume_equations = True
nbsphinx_allow_errors = True
nbsphinx_execute = "never"
nbsphinx_output_prompt = "Out[%s]:"

mathjax3_config = {
    "tex": {"tags": "ams", "useLabelIds": False},
}


autoapi_dirs = [str(Path("..", "..", "src"))]
autoapi_type = "python"
autoapi_generate_api_docs = True
autoapi_keep_files = False
autoapi_template_dir = "_templates/autoapi/python"

napoleon_include_init_with_doc = False


mathjax_path = "https://cdn.jsdelivr.net/npm/mathjax@2/MathJax.js?config=TeX-AMS-MML_HTMLorMML"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]



"""Functions for algebraic manipulation of axiom expressions.

@author: Andrés Corrada-Emmanuel."""

from collections.abc import Sequence

import sympy


def extract_coefficents(
    expr: sympy.UnevaluatedExpr, vars: Sequence[sympy.Symbol]
) -> Sequence[int]:
    """
    Extract vars coefficients from expr.

    Assumes that 'expr' is a linear equation.

    Parameters
    ----------
    expr : sympy.UnevaluatedExpr
        Linear expression.
    vars : Sequence[sympy.Symbol]
        Variables for which we have their coefficient in expr.

    Returns
    -------
    Sequence[int]
        The integer coefficients, following the same order as 'vars'.

    """
    coeff_dict = expr.as_coefficients_dict()
    zero = sympy.core.numbers.Zero()
    var_coeffs = tuple(int(sympy.N(coeff_dict.get(var, zero))) for var in vars)

    return var_coeffs


def extract_constant(expr: sympy.UnevaluatedExpr) -> int:
    """
    Extracts the constant term in 'expr.'

    Returns the value of the constant term in expr.

    Parameters
    ----------
    expr : sympy.UnevaluatedExpr
        An algebraic expression.

    Returns
    -------
    int
        The integer value of the constant term.

    """
    one = sympy.core.numbers.One()
    const = int(expr.as_coefficients_dict().get(one, 0))
    return const



"""@author: Andrés Corrada-Emmanuel."""

from itertools import combinations, product
from typing import Any
from types import MappingProxyType
from typing_extensions import Iterable, Literal, Mapping, Sequence, Union

import sympy

from ntqr import Labels
from ntqr.statistics import MClassifiersVariables


class MAxiomsIdeal:
    """
    Class for generating the axioms related an (M=m)-sized subsets
    of the classifiers.

    Each subset of the classifiers has a set of axioms, of size R,
    involving the marginalized decisions of that subset. For each
    value of the size of a subset, M, we can establish universal
    relations between counts of observed M-sized decision tuples
    and the counts of those tuples and smaller ones given true labels.

    The bottom of the M axioms ladder is M=1, the individual axioms.
    These carve out the evaluations for a given test taker given their
    marginalized response counts. This is a subset of the R-simplex,
    for an individual test taker, for any test of size Q. Thus, each
    test-taker in a group of N gets their own M=1 set of axioms.

    The M=2 axioms correspond to all possible pairs in a group of N
    test takers. These involve all the variables in the  M=1 axioms
    but now define a new set of axioms involving the pair response
    variables.

    In R-space, the space of integer response counts, all these
    ideals are linear equations. Thus, it may seem that calling them
    'ideals', while strictly true, is overkill. However, in P-space,
    these ideals are polynomials of degree 2 or higher.
    """

    def __init__(
        self,
        labels: Labels,
        classifiers: Sequence[str],
        m: int = 1,
    ):
        """
        The M=m axioms given answer labels and classifier labels.

        Parameters
        ----------
        labels : Labels
            The R possible label answers to the Q questions.
        classifiers : Sequence[str]
            Labels for indexing classifiers.
        m : int, optional
            M=m axioms index. The default is 1.

        Returns
        -------
        None.

        """
        self.labels = labels
        self.classifiers = classifiers
        self.m = m

        # We get all the variables associated with these classifiers
        # and reorganize them to map m_subset -> "vars" -> vars
        self.mvars = MClassifiersVariables(labels, classifiers)
        self._m_complex = {}
        for m_size in range(1, m + 1):
            for m_subset in combinations(classifiers, m_size):
                self._m_complex[m_subset] = {
                    "label_rvars": self.mvars.label_responses[m_subset],
                    "rvars": self.mvars.responses[m_subset],
                }

        self.all_agree_subs_dict = self.initialize_all_agree_subs_dict()

        # Now we compute the variety for all subsets of the classifiers of
        # size m.
        for m_subset in combinations(classifiers, m):

            match m:
                case 1:
                    axiomatic_ideal = self.m_one_ideal_agreement(m_subset)
                case 2:
                    axiomatic_ideal = self.m_two_ideal_agreement(m_subset)
                case 3:
                    ideal_cr = self.m_three_ideal_agreement_representation(
                        m_subset
                    )

                    axiomatic_ideal = {
                        label: sympy.simplify(
                            axiom.subs(self.all_agree_subs_dict)
                        )
                        for label, axiom in ideal_cr.items()
                    }
                case _:
                    raise ValueError(
                        "Only up to M=2 axiom ideals are currently supported."
                    )

            self._m_complex[m_subset]["axioms"] = axiomatic_ideal

        self.m_complex = MappingProxyType(self._m_complex)

    def m_one_ideal_agreement(
        self, classifier: tuple[str]
    ) -> Mapping[str, sympy.UnevaluatedExpr]:
        """
        The M=1 axioms ideal with agreement label response variables.

        The axioms in label response space are easiest to understand
        and prove when we use 'agreement' variables. Those are variables
        where all the classifiers are agreeing in their responses on the
        true label.

        Parameters
        ----------
        classifier : tuple[str]
            Classifier to use for variable subscripts.

        Returns
        -------
        m1_axioms_ideal :  Mapping[str, sympy.UnevaluatedExpr]
            A mapping from label to its corresponding m=1 axiom
            for the given classifier.

        """
        m1_responses = self.mvars.responses[classifier]
        m1_label_responses = self.mvars.label_responses[classifier]
        qs = self.mvars.qs

        m1_axioms_ideal = {
            l_true: sympy.simplify(
                -m1_label_responses[l_true][(l_true,)]
                + qs[l_true]
                # The single response terms
                - sum(
                    m1_responses[(l_error,)]
                    for l_error in self.labels
                    if l_error != l_true
                )
                + sum(
                    m1_label_responses[l_e2][(l_e1,)]
                    for l_e1 in self.labels
                    for l_e2 in self.labels
                    if (l_e1 != l_true) and (l_e2 != l_true)
                )
            )
            for l_true in self.labels
        }

        return m1_axioms_ideal

    def _m_one_ideal(
        self, labels: tuple[str], m_subset: tuple[str]
    ) -> Mapping[str, sympy.UnevaluatedExpr]:
        """
        Compute M=1 ideal expressed solely in terms of disagreement variables.

        This may need to be deprecated. The original rationale for this
        function is that all variables belong to some simplex and therefore
        we can always get rid of the variable corresponding to all of
        the test takers agreeing on the correct answer so as to save
        some computational load.

        The problem is that this form is more complex than the expression
        returned by self.m_one_ideal_agreement. This also made it more bug
        prone. Future development of NTQR will be based on the "natural"
        representation of the axioms indexed by the all-correct response
        variable. These are easier to write down from the top of one's head
        and easier to write mechanical proofs for them using more
        expressive algebraic systems like the Wolfram language.

        This function will be deprecated as it does the same computation
        as self.m_one_ideal_agreement.

        Parameters
        ----------
        labels : Sequence[str]
            Labels.
        m_subset : tuple[str]
            M=m sequence of classifiers.

        Returns
        -------
        axioms_by_label : Mapping[str->sympy.UnevaluatedExpr]
            M=m axioms indexed by label.

        """
        qs = self.mvars.qs
        responses = self._m_complex[m_subset]["rvars"]
        responses_by_label = self._m_complex[m_subset]["label_rvars"]

        axioms_by_label = {
            l_true: sympy.UnevaluatedExpr(
                sum(qs[label] for label in self.labels if label != l_true)
                + sum(
                    var
                    for var in responses_by_label[l_true]["errors"].values()
                )
                - sum(
                    responses[(label,)]
                    for label in self.labels
                    if label != l_true
                )
                - sum(
                    label_responses["errors"][(l_true,)]
                    for label, label_responses in responses_by_label.items()
                    if label != l_true
                )
            )
            for l_true in labels
        }

        return axioms_by_label

    def _m_two_ideal(
        self, pair: tuple[Any, Any]
    ) -> dict[str, sympy.UnevaluatedExpr]:
        """
        Construct the m=2 ideal.

        This is the 'errors' variables version. The one used for internal
        computation of the varieties. It has the drawback that it is hard
        to check or to write code for it.

        Parameters
        ----------
        pair : Sequence[Any, Any]
            Pair of classifiers.

        Returns
        -------
        m2_axioms_ideal : Mapping[label, sympy.UnevaluatedExpr]
            A mapping from label to its corresponding r-axiom.

        """
        labels = self.labels
        qs = self.mvars.qs
        m2_responses = self._m_complex[pair]["rvars"]
        m2_label_responses = self._m_complex[pair]["label_rvars"]

        # Now we have to build the variables for m1 decision
        # tuples.
        m1_responses = {
            m1: self._m_complex[m1]["rvars"] for m1 in combinations(pair, 1)
        }
        m1_label_responses = {
            m1: self._m_complex[m1]["label_rvars"]
            for m1 in combinations(pair, 1)
        }

        m2_axioms_ideal = {
            l_true: (
                sympy.simplify(
                    sum([qs[q_label] for q_label in qs if q_label != l_true])
                    + sum(
                        var
                        for var in m2_label_responses[l_true][
                            "errors"
                        ].values()
                    )
                    + sum(
                        var
                        for error_pair, var in m2_label_responses[l_true][
                            "errors"
                        ].items()
                        if (
                            (error_pair[0] != error_pair[1])
                            and (error_pair[0] != l_true)
                            and (error_pair[1] != l_true)
                        )
                    )
                    # The single response terms
                    - sum(
                        m1_responses[m1][(l_error,)]
                        for m1 in combinations(pair, 1)
                        for l_error in labels
                        if l_error != l_true
                    )
                    + sum(
                        m1_label_responses[m1][l_e1][(l_e2,)]
                        for m1 in combinations(pair, 1)
                        for l_e1 in labels
                        for l_e2 in labels
                        if (l_e1 != l_true)
                        and (l_e2 != l_true)
                        and (l_e1 != l_e2)
                    )
                    - sum(
                        m1_label_responses[m1][l_e1][(l_e2,)]
                        for m1 in combinations(pair, 1)
                        for l_e1 in labels
                        for l_e2 in labels
                        if ((l_e1 != l_true) and (l_e2 != l_e1))
                    )
                    # The m2 terms
                    + sum(
                        m2_responses[(l_error, l_error)]
                        for l_error in labels
                        if l_error != l_true
                    )
                    #
                    - sum(
                        m2_label_responses[l_e2][(l_e1, l_e1)]
                        for l_e1 in labels
                        for l_e2 in labels
                        if (l_e1 != l_true)
                        and (l_e1 != l_e2)
                        and (l_e2 != l_true)
                    )
                    + sum(
                        var
                        for l_e1 in self.labels
                        for var in m2_label_responses[l_e1]["errors"].values()
                        if (l_e1 != l_true)
                    )
                )
            )
            for l_true in labels
        }

        return m2_axioms_ideal

    def m_two_ideal_agreement(
        self, pair: tuple[Any, Any]
    ) -> Mapping[str, sympy.UnevaluatedExpr]:
        """
        Construct the M=2 algebraic ideal.

        The axioms in label response space are easiest to understand
        and prove when we use 'agreement' variables. Those are variables
        where all the classifiers are agreeing in their responses on the
        true label.

        Starting with M=2 this will be the only way the axioms will be
        encoded from now on. The representation with 'errors' variables
        only, the one needed for generalizable code, will be created with
        the straightforward transformation that gives the all agree on the
        true label as the number of questions of that label minus all the
        other possible responses.

        Parameters
        ----------
        pair : tuple[Any, Any]
            The classifier pair.

        Returns
        -------
        m2_axioms_ideal : Mapping[str, sympy.UnevaluatedExpr]
            Mapping from labels to its corresponding M=2 axiom.

        """
        labels = self.labels
        qs = self.mvars.qs
        m2_responses = self.mvars.responses[pair]
        m2_label_responses = self.mvars.label_responses[pair]

        # Now we have to build the variables for m1 decision
        # tuples.
        m1_responses = {
            m1: self.mvars.responses[m1] for m1 in combinations(pair, 1)
        }
        m1_label_responses = {
            m1: self.mvars.label_responses[m1] for m1 in combinations(pair, 1)
        }

        m2_axioms_ideal = {
            l_true: sympy.simplify(
                -m2_label_responses[l_true][(l_true, l_true)]
                + qs[l_true]
                # The single response terms
                - sum(
                    m1_responses[m1][(l_error,)]
                    for m1 in combinations(pair, 1)
                    for l_error in labels
                    if l_error != l_true
                )
                + sum(
                    m1_label_responses[m1][l_e2][(l_e1,)]
                    for m1 in combinations(pair, 1)
                    for l_e1 in labels
                    for l_e2 in labels
                    if (l_e1 != l_true) and (l_e2 != l_true)
                )
                # The m2 terms
                + sum(
                    m2_responses[(l_error, l_error)]
                    for l_error in labels
                    if l_error != l_true
                )
                - sum(
                    m2_label_responses[l_e2][(l_e1, l_e1)]
                    for l_e1 in self.labels
                    for l_e2 in self.labels
                    if (l_e1 != l_true) and (l_e2 != l_true)
                )
                + sum(
                    m2_label_responses[l_true]["errors"][(l_e1, l_e2)]
                    for l_e1 in self.labels
                    for l_e2 in self.labels
                    if (l_e1 != l_e2) and (l_e1 != l_true) and (l_e2 != l_true)
                )
            )
            for l_true in labels
        }

        return m2_axioms_ideal

    def m_three_ideal_agreement_representation(
        self, trio: tuple[str | int, str | int, str | int]
    ) -> Mapping[str, sympy.UnevaluatedExpr]:
        """
        Construct the M=3 algebraic ideal with agreement variables.

        Parameters
        ----------
        trio : tuple[str | int, str | int]
            Any three classifiers.

        Returns
        -------
        m2_axioms_ideal : Mapping[str, sympy.UnevaluatedExpr]
            Dictionary from label to its corresponding axiom.

        """
        labels = self.labels
        qs = self.mvars.qs
        m3_responses = self.mvars.responses[trio]
        m3_label_responses = self.mvars.label_responses[trio]

        # Now we have to build the variables for m1 decision
        # tuples.
        m1_responses = {
            m1: self.mvars.responses[m1] for m1 in combinations(trio, 1)
        }
        m1_label_responses = {
            m1: self.mvars.label_responses[m1] for m1 in combinations(trio, 1)
        }

        # Now we have to build the variables for m2 decision
        # tuples.
        m2_responses = {
            m1: self.mvars.responses[m1] for m1 in combinations(trio, 2)
        }
        m2_label_responses = {
            m1: self.mvars.label_responses[m1] for m1 in combinations(trio, 2)
        }

        m3_axioms_ideal = {
            l_true: sympy.simplify(
                -m3_label_responses[l_true][(l_true, l_true, l_true)]
                + qs[l_true]
                # The single response terms
                - sum(
                    m1_responses[m1][(l_error,)]
                    for m1 in combinations(trio, 1)
                    for l_error in labels
                    if l_error != l_true
                )
                + sum(
                    m1_label_responses[m1][l_e2][(l_e1,)]
                    for m1 in combinations(trio, 1)
                    for l_e1 in labels
                    for l_e2 in labels
                    if (l_e1 != l_true) and (l_e2 != l_true)
                )
                # The m2 terms
                + sum(
                    m2_responses[pair][(l_error, l_error)]
                    for pair in combinations(trio, 2)
                    for l_error in labels
                    if l_error != l_true
                )
                - sum(
                    m2_label_responses[pair][l_e2][(l_e1, l_e1)]
                    for pair in combinations(trio, 2)
                    for l_e1 in self.labels
                    for l_e2 in self.labels
                    if (l_e1 != l_true) and (l_e2 != l_true)
                )
                # The m3 terms
                - sum(
                    m3_responses[(l_error, l_error, l_error)]
                    for l_error in labels
                    if l_error != l_true
                )
                + sum(
                    m3_label_responses[l_e2][(l_e1, l_e1, l_e1)]
                    for l_e1 in self.labels
                    for l_e2 in self.labels
                    if (l_e1 != l_true) and (l_e2 != l_true)
                )
                # Two do not agree
                + sum(
                    m3_label_responses[l_true]["errors"][(l_e1, l_e2, l_e3)]
                    for l_e1 in self.labels
                    for l_e2 in self.labels
                    for l_e3 in self.labels
                    if len(set((l_e1, l_e2, l_e3, l_true))) == 3
                )
                # Three do not agree
                + 2
                * sum(
                    m3_label_responses[l_true]["errors"][(l_e1, l_e2, l_e3)]
                    for l_e1 in self.labels
                    for l_e2 in self.labels
                    for l_e3 in self.labels
                    if len(set((l_e1, l_e2, l_e3, l_true))) == 4
                )
            )
            for l_true in labels
        }

        return m3_axioms_ideal

    def initialize_all_agree_subs_dict(self) -> None:
        """
        Initialize the substitution dictionary for all correct responses.

        Computations of the label response simplex variables is easiest
        when we represent all agree on the correct label in terms of
        the disagreeing label responses variables. This function
        initializes the substitution dictionary that can be used to
        turn any expression in terms of label response variables into
        one that only uses disagreement ones.

        Returns
        -------
        None
            The self.all_correct_subs_dict is initialized.

        """
        subs_dict = {}
        for m_subset, m_subset_dict in self._m_complex.items():
            for l_true in self.labels:
                var = m_subset_dict["label_rvars"][l_true][
                    tuple([l_true for i in range(len(m_subset))])
                ]
                subs_dict[var] = self.mvars.qs[l_true] - sum(
                    m_subset_dict["label_rvars"][l_true]["errors"].values()
                )

        return subs_dict
